﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.3
@EndOfDesignText@
Sub Class_Globals
	Private xui As XUI
	Private ImageViewPanel As B4XView
	Private ImageView1 As B4XView
	Public Utils As GameUtils
	Public Rows As List
	Public BlockSize As Int = 30
	Public BlocksPerRow As Int = 10
	Public NumberOfRows As Int = 20
	Type PieceData (Pattern As List, BlockBC As BitmapCreator, Orientations As Int, CenterX As Int, CenterY As Int)
	Private PDs As List
	Private Background1 As Background
	Public IntervalBetweenMovements As Int
	Private ScoreLabel1 As ScoreLabel
	Private btnPause As B4XView
	Private lblLevel As B4XView
	Private level As Int = 1
	Private GameOverState As Boolean
	Private btnSound As B4XView
	Public PlaySound As Boolean
	Private MessageToSet As String
	Private imgNextPiece As B4XView
	Private NextPieceType As Int
	Private NextPieceBC As BitmapCreator
	Private lblNext As Label
	#if B4i
	Private FeedbackGenerator As NativeObject
	#End If
End Sub

Public Sub Initialize (Parent As B4XView)
	Parent.LoadLayout("1")
	ImageViewPanel.LoadLayout("ImageView")
	Utils.Initialize(Me, BlockSize * BlocksPerRow, BlockSize * NumberOfRows , ImageView1)
	PDs.Initialize
	LoadPieces
	NextPieceBC.Initialize(BlockSize * 4 + 10, BlockSize * 4 + 10)
	ResetState
	
	Dim fnt As B4XFont
	#if B4J
	Dim fx As JFX
	fnt = fx.LoadFont(File.DirAssets, "neutronium.ttf", 30)
	#else if B4A
	fnt = xui.CreateFont(Typeface.LoadFromAssets("neutronium.ttf"), 20)
	#else if B4i
	fnt = xui.CreateFont(Font.CreateNew2("Neutronium", 20), 20)
	FeedbackGenerator = FeedbackGenerator.Initialize("UIImpactFeedbackGenerator").RunMethod("alloc", Null).RunMethod("initWithStyle:", Array(0)) 'light
	#End If
	For Each lbl As B4XView In Array(ScoreLabel1.Base.GetView(0), lblLevel, lblNext)
		#if B4J
		Dim jlbl As Label = lbl
		jlbl.Style = "" 'to avoid conflicts between the CSS settings and the custom font
		#else if B4i
		Dim ilbl As Label = lbl
		ilbl.Multiline = True
		#End If
		lbl.Font = fnt
		lbl.SetTextAlignment("CENTER", "CENTER")
		lbl.TextColor = xui.Color_White
	Next
	ScoreLabel1.Text = "Score" & CRLF
	btnSound_Click

End Sub

Public Sub Start
	Main.DownDown = False
	Main.LeftDown = False
	Main.RightDown = False
	Main.UpDown = False
	Utils.Start
End Sub

Private Sub ResetState
	GameOverState = False
	Utils.Reset
	Rows.Initialize
	For i = 0 To NumberOfRows - 1
		Dim row(BlocksPerRow) As Boolean
		Rows.Add(row)
	Next
	Background1.Initialize(Utils.CreateSprite(Background1))
	Background1.MySprite.SkipBlending = True
	Background1.MySprite.Position.x = Utils.ViewRect.CenterX
	Background1.MySprite.Position.y = Utils.ViewRect.CenterY
	NextPieceType = Rnd(0, PDs.Size)
	CreateNewPiece
	ScoreLabel1.SetValueNow(0)
	UpdateLevel(1)
End Sub

Private Sub UpdateLevel (lvl As Int)
	level = lvl
	lblLevel.Text = "Level" & CRLF & lvl
	IntervalBetweenMovements = 21 - level
End Sub

Private Sub btnPause_Click
	SetPaused (Utils.IsRunning)
End Sub

Public Sub SetPaused (paused As Boolean)
	If paused = Not(Utils.IsRunning) Then Return
	If Utils.IsRunning Then
		Utils.Stop
		btnPause.Text = Chr(0xF04B)
		If GameOverState = False Then
			ShowMessage("Paused")
		End If
	Else
		btnPause.Text = Chr(0xF04C)
		If GameOverState Then ResetState
		Start
	End If
End Sub

Sub btnSound_Click
	PlaySound = Not(PlaySound)
	If PlaySound Then
		btnSound.Text = Chr(0xF028)
	Else
		btnSound.Text = Chr(0xF026)
	End If
End Sub

Private Sub LoadPieces
	Dim blockbmp As B4XBitmap = xui.LoadBitmapResize(File.DirAssets, "block.png", BlockSize, BlockSize, False)
	Dim bcblock As BitmapCreator = Utils.BitmapToBitmapCreator(blockbmp, True)
	Dim m As Map = File.ReadMap(File.DirAssets, "pieces.txt")
	Dim i As Int = 1
	Dim rb As RegexBuilder
	rb.Initialize.AppendEscaped("(").StartCapture.AppendAnyBut(Array(",")).AppendAtLeastOne.EndCapture.AppendEscaped(",")
	rb.StartCapture.AppendAnyBut(Array(")")).AppendAtLeastOne.EndCapture.AppendEscaped(")")
	Do While m.ContainsKey($"P${i}.Color"$)
		Dim pd As PieceData
		pd.Initialize
		'the full value as an unsigned number is too large to fit in a signed int so we need to add the alpha level ourselves.
		Dim clr As Int = 0xff000000 + Bit.ParseInt(m.Get($"P${i}.Color"$), 16)
		pd.BlockBC = Utils.GreyscaleToColor(bcblock, clr)
		pd.Orientations = m.Get($"P${i}.Orientations"$)
		pd.Pattern.Initialize
		Dim match As Matcher = Regex.Matcher(rb.Pattern, m.Get($"P${i}.Pattern"$))
		Do While match.Find
			pd.Pattern.Add(Array As Int(match.Group(1), match.Group(2)))
		Loop
		match = Regex.Matcher(rb.Pattern, m.Get($"P${i}.Center"$))
		match.Find
		pd.CenterX = match.Group(1)
		pd.CenterY = match.Group(2)
		PDs.Add(pd)
		i = i + 1
	Loop
End Sub

Public Sub IsTaken (x As Int, y As Int) As Boolean
	Dim row() As Boolean = Rows.Get(y)
	Return row(x)
End Sub

Public Sub SetTaken (x As Int, y As Int, state As Boolean)
	Dim row() As Boolean = Rows.Get(y)
	row(x) = state
End Sub

Public Sub SpriteRemoved (Sprite1 As Sprite)

End Sub

Public Sub PieceStoppedMoving (piece1 As Piece)
	Dim dt As DrawTask = piece1.MySprite.ToDrawTask
	Background1.MySprite.Frame.DrawBitmapCreator(dt.Source, dt.SrcRect, dt.TargetX, dt.TargetY, dt.SkipBlending)
	Dim RemovedRows As List
	Dim i As Int = 0
	Do While i < NumberOfRows
		Dim row() As Boolean = Rows.Get(i)
		If IsFullRow(row) Then
			Rows.RemoveAt(i)
			Dim b(BlocksPerRow) As Boolean
			Rows.InsertAt(0, b)
			If RemovedRows.IsInitialized = False Then RemovedRows.Initialize
			RemovedRows.Add(i)
			ScoreLabel1.Value = ScoreLabel1.Value + 100
			If ScoreLabel1.Value Mod 200 = 0 Then
				UpdateLevel(level + 1)
			End If
		End If
		i = i + 1
	Loop
	If RemovedRows.IsInitialized Then 
		CallSub(Main, "PlayRow")
		Background1.RemoveRows(RemovedRows)
	End If
	CreateNewPiece
	DebugCurrentState	
End Sub

Private Sub GameOver
	CallSub(Main, "PlayGameOver")
	GameOverState = True
	SetPaused(True)
	ShowMessage("Game Over!!!")
End Sub

'game must be paused when called.
Private Sub ShowMessage (t As String)
	MessageToSet = t
	Dim bc As BitmapCreator = Utils.MainBC
	Dim cvs As B4XCanvas  = Utils.CreateCanvas(bc.mWidth, 200)
	cvs.DrawRect(cvs.TargetRect, 0x80000000, True, 0)
	cvs.DrawText(t, bc.mWidth / 2, 120, xui.CreateFont2(ScoreLabel1.Base.GetView(0).Font, 40 / xui.Scale), xui.Color_White, _
		"CENTER")
	cvs.Invalidate
	Sleep(100)
	MessageToSet = ""
	If Utils.IsRunning = True Then Return
	Dim r As B4XRect
	r.Initialize(0, bc.mHeight / 2 - cvs.TargetRect.Height / 2, bc.mWidth, bc.mHeight / 2 + cvs.TargetRect.Height / 2)
	bc.DrawBitmap(cvs.CreateBitmap, r, False)
	Utils.SetBitmapWithFitOrFill(Utils.mTargetView, bc.Bitmap)
End Sub

Public Sub Activity_Resume
	'the pause message will not appear if the user pressed on the home button or power button
	'we need to set it again.
	If MessageToSet <> "" And Utils.IsRunning = False Then
		ShowMessage(MessageToSet)
	End If
End Sub

Private Sub DebugCurrentState
	Log("*********************")
	Dim sb As StringBuilder
	sb.Initialize
	For Each row() As Boolean In Rows
		For Each c As Boolean In row
			If c Then sb.Append("X") Else sb.Append("o")
		Next
		sb.Append(CRLF)
	Next
	Log(sb)
End Sub

Private Sub IsFullRow(row() As Boolean) As Boolean
	For Each b As Boolean In row
		If b = False Then Return False
	Next
	Return True
End Sub

Public Sub CreateNewPiece
	Dim piece1 As Piece
	If piece1.Initialize(Utils.CreateSprite(piece1), PDs.Get(NextPieceType)) = False Then
		GameOver
	Else
		NextPieceType = Rnd(0, PDs.Size)
		DrawNextPiece(PDs.Get(NextPieceType))	
	End If
End Sub

Private Sub DrawNextPiece (pd As PieceData)
	NextPieceBC.FillRect(xui.Color_Transparent, NextPieceBC.TargetRect)
	Dim ActualBounds As B4XRect
	ActualBounds.Initialize(1000, 1000, 0, 0) 'ignore
	'First we find the actual bounds of the piece as it is not centered.
	For Each xy() As Int In pd.Pattern
		Dim left As Int = (pd.CenterX + xy(0)) * BlockSize 
		Dim top As Int = (pd.CenterY + xy(1)) * BlockSize
		ActualBounds.Left = Min(ActualBounds.Left, left)
		ActualBounds.Top = Min(ActualBounds.Top, top)
		ActualBounds.Right = Max(ActualBounds.Right, left + BlockSize)
		ActualBounds.Bottom = Max(ActualBounds.Bottom, top + BlockSize)
	Next
	Dim OffsetX As Int = NextPieceBC.mWidth / 2 - ActualBounds.CenterX
	Dim OffsetY As Int = NextPieceBC.mHeight / 2 - ActualBounds.CenterY
	For Each xy() As Int In pd.Pattern
		Dim left As Int = (pd.CenterX + xy(0)) * BlockSize + OffsetX
		Dim top As Int = (pd.CenterY + xy(1)) * BlockSize + OffsetY
		NextPieceBC.DrawBitmapCreator(pd.BlockBC, pd.BlockBC.TargetRect, left, top, True)
	Next
	Utils.SetBitmapWithFitOrFill(imgNextPiece, NextPieceBC.Bitmap)
End Sub

Public Sub Tick (GameStep As GameStep)
	ScoreLabel1.Tick
End Sub

Public Sub AfterFirstLoop As ResumableSub
	Return True
End Sub



#if B4A or B4i

'Sub ImageViewPanel_Click
'	Main.UpDown = True
'	PerformHapticFeedback(Sender)
'End Sub

Sub lblDown_Click
	Main.DownDown = True
	PerformHapticFeedback(Sender)
End Sub

Sub lblRight_Click
	Main.RightDown = True
	PerformHapticFeedback(Sender)
End Sub

Sub lblRotate_Click
	Main.UpDown = True
	PerformHapticFeedback(Sender)
End Sub

Sub lblLeft_Click
	Main.LeftDown = True
	PerformHapticFeedback(Sender)
End Sub

Private Sub PerformHapticFeedback (view As Object)
	#if B4A
	Dim jo As JavaObject = view
	jo.RunMethod("performHapticFeedback", Array(1))
	#Else
	FeedbackGenerator.RunMethod("impactOccurred", Null)
	#end if
End Sub
#end if